"""
Implementation for the Seymour RS232 protocol version 01.
"""

from __future__ import annotations

import enum
import re

import attrs

FRAME_START = b"["
FRAME_END = b"]"
PROTOCOL_VERSION = b"01"


class SingleByte(enum.StrEnum):
    """
    Base class for enums that represent a 1-byte code.
    """

    @classmethod
    def regex(cls, n: int | None = None) -> bytes:
        """
        Regular expression that parses exactly one code value into a capture group named after the class.
        """
        all_commands = "".join(code.value for code in cls)
        capture_group_suffix = str(n) if n is not None else ""
        return rf"(?P<{cls.__name__}{capture_group_suffix}>[{all_commands}])".encode("ascii")


class CommandCode(SingleByte):
    MOVE_OUT = "O"
    MOVE_IN = "I"
    MOVE_TO_RATIO = "M"
    HOME = "A"
    HALT = "H"
    CALIBRATE = "C"
    STATUS = "S"
    POSITIONS = "P"
    UPDATE_RATIO = "U"
    READ_SYSTEM_INFO = "Y"
    READ_SETTINGS = "R"
    CLEAR_SETTINGS = "X"
    DIAGNOSTICS = "@"


class MotorID(SingleByte):
    TOP = "T"
    BOTTOM = "B"
    LEFT = "L"
    RIGHT = "R"
    VERTICAL = "V"
    HORIZONTAL = "H"
    ALL = "A"


class MovementCode(enum.StrEnum):
    UNTIL_LIMIT = ""
    JOG = "J"  # smallest increment (<0.1%)
    MOVE = "M"  # 1% of the motor's range (ignored if not calibrated)


class StatusCode(SingleByte):
    STOPPED_AT_RATIO = "P"
    MOVING_TO_RATIO = "M"
    HALTED = "H"
    HOMING = "A"
    CALIBRATING = "C"
    MOVING_OUTWARD = "O"
    MOVING_INWARD = "I"
    ERROR = "E"


@attrs.frozen
class Ratio:
    id: str = attrs.field()

    @id.validator
    def check(self, _attribute: attrs.Attribute[str], value: str) -> None:
        if len(value) != 3:
            raise ValueError("ratio_id must be three digits long")
        if not value.isdigit():
            raise ValueError("ratio_id must be numeric")

    @classmethod
    def regex(cls) -> bytes:
        """
        Regular expression that parses a ratio ID into a capture group named 'Ratio'.
        """
        return rb"(?P<Ratio>[0-9]{3})"


@attrs.frozen
class MaskStatus:
    code: StatusCode
    ratio: Ratio | None = None


@attrs.frozen
class MaskPosition:
    motor_id: MotorID
    position_pct: float


@attrs.frozen
class Serial:
    model_code: str
    month: int
    year: int
    production_number: str

    def to_serial_number(self) -> str:
        return f"{self.model_code}-{self.month:02d}{self.year:02d}-{self.production_number}"

    PARSER_RE = re.compile(
        rb"(?P<model_code>.{2})-(?P<month>[0-9]{2})(?P<year>[0-9]{2})-(?P<production_number>.{5})"
    )

    @staticmethod
    def from_serial_number(serial_number: bytes) -> Serial:
        """
        Parse a serial number bytestring into a Serial object.

        The expected format is "XX-MMYY-PPPPP", where XX is the model code and PPPPP is the
        production number. The hyphens are required.
        """
        match = Serial.PARSER_RE.fullmatch(serial_number)
        if not match:
            raise ValueError(f"Malformed serial number: {serial_number!r}")

        model_code = match.group("model_code").decode("ascii")
        month = int(match.group("month"))
        year = int(match.group("year"))
        production_number = match.group("production_number").decode("ascii")
        return Serial(
            model_code=model_code,
            month=month,
            year=year,
            production_number=production_number,
        )


@attrs.frozen
class SystemInfo:
    screen_model: str
    width_inches: float
    height_inches: float
    serial_number: Serial
    mask_ids: list[MotorID]


@attrs.frozen
class RatioSetting:
    ratio: Ratio
    label: str
    width_inches: float
    height_inches: float
    motor_positions_pct: list[float]
    motor_adjustments_pct: list[float]

    @staticmethod
    def from_response_entry(entry: bytes, num_motors: int) -> RatioSetting:
        """
        Parse a ratio setting info string into a RatioSetting object.

        The expected format is:
        - 3 bytes: ratio ID
        - 8 bytes: label (ASCII, space-padded)
        - 6 bytes: width in inches (ASCII float)
        - 6 bytes: height in inches (ASCII float)
        - For each motor:
          - 4 bytes: position percentage (ASCII float)
        - For each motor:
          - 4 bytes: position adjustment percentage (ASCII float)
        """
        base_pattern = Ratio.regex() + rb"(?P<label>.{8})(?P<width>[0-9.]{6})(?P<height>[0-9.]{6})"

        pattern = base_pattern + _ASCII_PCT_RE * num_motors * 2

        match = re.fullmatch(pattern, entry)
        if not match:
            raise ValueError(f"Malformed ratio setting entry for pattern {pattern!r}: {entry!r}")

        ratio_id = match.group("Ratio").decode("ascii")
        label = match.group("label").decode("ascii").strip()
        width = float(match.group("width"))
        height = float(match.group("height"))

        # parse the motor positions and adjustments
        motor_positions_pct: list[float] = []
        motor_adjustments_pct: list[float] = []
        position_group_start = 5  # after ratio, label, width, height
        for i in range(num_motors):
            pos_str = match.group(position_group_start + i)
            adj_str = match.group(position_group_start + num_motors + i)
            motor_positions_pct.append(float(pos_str))
            motor_adjustments_pct.append(float(adj_str))

        return RatioSetting(
            ratio=Ratio(ratio_id),
            label=label,
            width_inches=width,
            height_inches=height,
            motor_positions_pct=motor_positions_pct,
            motor_adjustments_pct=motor_adjustments_pct,
        )

    @staticmethod
    def expected_entry_length(num_motors: int) -> int:
        """
        Calculate the expected length of a ratio setting entry for the given number of motors.
        """
        # 3 (ratio ID) + 8 (label) + 6 (width) + 6 (height) + N * 4 (positions) + N * 4 (adjustments)
        return 3 + 8 + 6 + 6 + num_motors * 4 + num_motors * 4


class DiagnosticCommand(SingleByte):
    DEBUG_LOG = "D"


class DiagnosticOption(enum.StrEnum):
    LIST_FS = "00"
    LIST_SETTINGS_JSON = "10"
    LIST_SYSTEM_JSON = "20"


def _frame(payload: str) -> bytes:
    return FRAME_START + PROTOCOL_VERSION + payload.encode("ascii") + FRAME_END


def encode_move_out(motor_id: MotorID, movement: MovementCode) -> bytes:
    return _frame(f"{CommandCode.MOVE_OUT}{motor_id}{movement}")


def encode_move_in(motor_id: MotorID, movement: MovementCode) -> bytes:
    return _frame(f"{CommandCode.MOVE_IN}{motor_id}{movement}")


def encode_move_ratio(ratio: Ratio) -> bytes:
    return _frame(f"{CommandCode.MOVE_TO_RATIO}{ratio.id}")


def encode_home(motor_id: MotorID) -> bytes:
    return _frame(f"{CommandCode.HOME}{motor_id}")


def encode_halt(motor_id: MotorID) -> bytes:
    return _frame(f"{CommandCode.HALT}{motor_id}")


def encode_calibrate(motor_id: MotorID) -> bytes:
    return _frame(f"{CommandCode.CALIBRATE}{motor_id}")


def encode_status() -> bytes:
    return _frame(CommandCode.STATUS)


def encode_positions() -> bytes:
    return _frame(CommandCode.POSITIONS)


def encode_update_ratio(ratio: Ratio) -> bytes:
    return _frame(f"{CommandCode.UPDATE_RATIO}{ratio.id}")


def encode_clear_settings(ratio: Ratio | None) -> bytes:
    id = ratio.id if ratio is not None else ""
    return _frame(f"{CommandCode.CLEAR_SETTINGS}{id}")


def encode_read_sysinfo() -> bytes:
    return _frame(CommandCode.READ_SYSTEM_INFO)


def encode_read_settings() -> bytes:
    return _frame(CommandCode.READ_SETTINGS)


def encode_diagnostics(option: DiagnosticOption) -> bytes:
    return _frame(f"{CommandCode.DIAGNOSTICS}{DiagnosticCommand.DEBUG_LOG}{option}")


def _frame_re(regex_body: bytes) -> re.Pattern[bytes]:
    regex = b"^\\" + FRAME_START + PROTOCOL_VERSION + regex_body + b"\\" + FRAME_END + b"$"
    return re.compile(regex)


_STATUS_RE = _frame_re(StatusCode.regex() + Ratio.regex() + rb"?")


def decode_status(raw: bytes) -> MaskStatus:
    match = _STATUS_RE.match(raw)
    if not match:
        raise ValueError(f"Malformed status response: {raw!r}")
    status_code = match.group("StatusCode").decode("ascii")
    maybe_ratio_str = match.group("Ratio")
    ratio = Ratio(maybe_ratio_str.decode("ascii")) if maybe_ratio_str is not None else None
    return MaskStatus(code=StatusCode(status_code), ratio=ratio)


_POSITIONS_RE = _frame_re(rb"(?P<num_motors>[1-4])(?P<entries>.+)")
_ASCII_PCT_RE = rb"([0-9.\-]{4})"


def decode_positions(raw: bytes) -> list[MaskPosition]:
    match = _POSITIONS_RE.match(raw)
    if not match:
        raise ValueError(f"Malformed positions response: {raw!r}")
    num_motors = int(match.group("num_motors"))
    entries_str = match.group("entries")

    pattern = b"".join(MotorID.regex(i) + _ASCII_PCT_RE for i in range(num_motors))
    match = re.fullmatch(pattern, entries_str)
    if not match:
        raise ValueError(f"Malformed motor position entries: {entries_str!r}")
    if len(match.groups()) != num_motors * 2:
        raise ValueError(f"Expected {num_motors} motor entries, got: {entries_str!r}")
    positions: list[MaskPosition] = []
    for i in range(num_motors):
        motor_id_str = match.group(1 + i * 2).decode("ascii")
        position_str = match.group(2 + i * 2).decode("ascii")
        positions.append(
            MaskPosition(
                motor_id=MotorID(motor_id_str),
                position_pct=float(position_str),
            )
        )
    return positions


_SYSTEM_INFO_RE = _frame_re(
    rb"(?P<model>.{20})(?P<width>[0-9.]{6})(?P<height>[0-9.]{6})(?P<serial>.{13})(?P<mask_ids>[LRTB]{1,5})"
)


def decode_system_info(raw: bytes) -> SystemInfo:
    match = _SYSTEM_INFO_RE.match(raw)
    if not match:
        raise ValueError(f"Malformed system info response: {raw!r}")
    model = match.group("model").decode("ascii").strip()

    width = float(match.group("width"))
    height = float(match.group("height"))

    serial_str = match.group("serial")
    serial = Serial.from_serial_number(serial_str)

    mask_ids_str = match.group("mask_ids").decode("ascii")
    mask_ids = [MotorID(c) for c in mask_ids_str]

    return SystemInfo(
        screen_model=model,
        width_inches=width,
        height_inches=height,
        serial_number=serial,
        mask_ids=mask_ids,
    )


_SETTINGS_RE = _frame_re(rb"(?P<num_motors>[0-9])(?P<num_ratios>[0-9]{2})(?P<entries>.+)")


def decode_settings(raw: bytes) -> list[RatioSetting]:
    match = _SETTINGS_RE.match(raw)
    if not match:
        raise ValueError(f"Malformed settings response: {raw!r}")
    num_motors = int(match.group("num_motors"))
    num_ratios = int(match.group("num_ratios"))
    entries_str = match.group("entries")

    expected_length_per_entry = RatioSetting.expected_entry_length(num_motors)
    if len(entries_str) != expected_length_per_entry * num_ratios:
        raise ValueError(
            f"Expected {num_ratios} ratio entries of length {expected_length_per_entry}, "
            f"got total length {len(entries_str)}: {entries_str!r}"
        )

    settings: list[RatioSetting] = []
    for i in range(num_ratios):
        entry_bytes = entries_str[
            i * expected_length_per_entry : (i + 1) * expected_length_per_entry
        ]
        setting = RatioSetting.from_response_entry(entry_bytes, num_motors)
        settings.append(setting)

    return settings



"""Transport abstraction: TCPTransport (for IP2SL) + SerialTransport (local serial)."""

from __future__ import annotations

import asyncio
from abc import abstractmethod

from .exceptions import SeymourTransportError

SEYMOUR_BAUD_RATE = 115200
ITACH_SERIAL_PORT = 4999


class SeymourTransport:
    """Abstract transport used by the SeymourClient."""

    def __init__(self) -> None:
        self.reader: asyncio.StreamReader | None = None
        self.writer: asyncio.StreamWriter | None = None
        self._lock = asyncio.Lock()

    @abstractmethod
    async def connect(self) -> None:
        raise NotImplementedError

    async def close(self) -> None:
        try:
            if self.writer:
                self.writer.close()
                await self.writer.wait_closed()
        finally:
            self.writer = None
            self.reader = None

    async def send(self, data: bytes) -> None:
        """Send one data frame to transport, blocking until the complete frame is sent."""
        assert data.startswith(b"["), "Frame must start with '['"
        assert data.endswith(b"]"), "Frame must end with ']'"

        if not self.writer:
            raise SeymourTransportError("Not connected")

        async with self._lock:
            self.writer.write(data)
            await self.writer.drain()

    async def receive(self) -> bytes:
        """
        Get one data frame from transport, blocking until frame end (']') is received.

        Note: no timeout here; caller should use structured concurrency if desired.
        """
        if not self.reader:
            raise SeymourTransportError("Not connected")

        try:
            data = await asyncio.wait_for(self.reader.readuntil(b"]"), timeout=None)
            return data
        except asyncio.IncompleteReadError as exc:
            raise SeymourTransportError("Connection closed") from exc


class TCPTransport(SeymourTransport):
    def __init__(self, host: str, port: int = ITACH_SERIAL_PORT) -> None:
        super().__init__()
        self.host = host
        self.port = port

    async def connect(self) -> None:
        self.reader, self.writer = await asyncio.open_connection(self.host, self.port)


class SerialTransport(SeymourTransport):
    def __init__(self, port: str, baudrate: int = SEYMOUR_BAUD_RATE) -> None:
        super().__init__()
        self.port = port
        self.baudrate = baudrate

    async def connect(self) -> None:
        try:
            import serial_asyncio
        except Exception as exc:
            raise SeymourTransportError("serial_asyncio not installed") from exc

        self.reader, self.writer = await serial_asyncio.open_serial_connection(
            url=self.port, baudrate=self.baudrate
        )



"""Main async client for interacting with the Seymour controller."""

from __future__ import annotations

import asyncio
import logging
import time

from tenacity import (
    AsyncRetrying,
    after_log,
    before_sleep_log,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from . import protocol
from .exceptions import SeymourConnectionError, SeymourProtocolError, SeymourTransportError
from .transport import SeymourTransport

_LOGGER = logging.getLogger(__name__)

DEFAULT_MAX_RETRIES = 3
DEFAULT_REQUEST_TIMEOUT = 10.0
HEALTH_CHECK_INTERVAL = 90.0
HEALTH_CHECK_TIMEOUT = 3.0


class SeymourClient:
    def __init__(
        self,
        transport: SeymourTransport,
        max_retries: int = DEFAULT_MAX_RETRIES,
        request_timeout: float = DEFAULT_REQUEST_TIMEOUT,
    ):
        self.transport = transport
        self._lock = asyncio.Lock()
        self._connected = False
        self.max_retries = max_retries
        self.request_timeout = request_timeout
        self._last_successful_operation = 0.0
        self._health_check_task: asyncio.Task[None] | None = None

    async def _connect_with_retry(self) -> None:
        """Connect with tenacity-managed retry logic."""
        retryer = AsyncRetrying(
            stop=stop_after_attempt(self.max_retries),
            wait=wait_exponential(multiplier=1, min=1, max=30),
            retry=retry_if_exception_type((SeymourTransportError, ConnectionError, OSError)),
            before_sleep=before_sleep_log(_LOGGER, logging.WARNING),
            after=after_log(_LOGGER, logging.INFO),
        )

        async for attempt in retryer:
            with attempt:
                await self.transport.connect()
                self._connected = True
                self._last_successful_operation = time.time()

    async def connect(self) -> None:
        async with self._lock:
            if self._connected:
                return

            try:
                await self._connect_with_retry()
            except Exception as exc:
                self._connected = False
                raise SeymourConnectionError("Failed to connect after all retries") from exc

    async def close(self) -> None:
        async with self._lock:
            # Stop health checking
            if self._health_check_task and not self._health_check_task.done():
                self._health_check_task.cancel()
                try:
                    await self._health_check_task
                except asyncio.CancelledError:
                    pass
                self._health_check_task = None

            # Close transport
            if self._connected:
                await self.transport.close()
                self._connected = False
                _LOGGER.info("Closed Seymour transport")

    async def start_health_monitoring(self, interval: float = HEALTH_CHECK_INTERVAL) -> None:
        """Start background health monitoring for long-running applications like Home Assistant."""
        if self._health_check_task and not self._health_check_task.done():
            return  # Already running

        self._health_check_task = asyncio.create_task(self._health_check_loop(interval))
        _LOGGER.info("Started health monitoring (interval=%.1fs)", interval)

    async def _health_check_loop(self, interval: float) -> None:
        """Background task that periodically checks connection health."""
        while True:
            try:
                await asyncio.sleep(interval)

                # Skip health check if we've had recent successful operations
                if time.time() - self._last_successful_operation < interval / 2:
                    continue

                # Perform lightweight health check
                if self._connected:
                    try:
                        await self._health_check()
                        _LOGGER.debug("Health check passed")
                    except Exception as exc:
                        _LOGGER.warning("Health check failed, marking disconnected: %s", exc)
                        self._connected = False

            except asyncio.CancelledError:
                _LOGGER.debug("Health monitoring cancelled")
                break
            except Exception as exc:
                _LOGGER.error("Unexpected error in health monitoring: %s", exc)
                await asyncio.sleep(interval)  # Continue monitoring despite errors

    async def _health_check(self) -> None:
        """Perform a lightweight health check by requesting status."""
        # Use a shorter timeout for health checks
        original_timeout = self.request_timeout
        self.request_timeout = min(HEALTH_CHECK_TIMEOUT, original_timeout)
        try:
            await self.get_status()
        finally:
            self.request_timeout = original_timeout

    @property
    def is_connected(self) -> bool:
        """Check if the client believes it's connected."""
        return self._connected

    @property
    def connection_stats(self) -> dict[str, float]:
        """Get connection statistics for monitoring."""
        return {
            "last_successful_operation": self._last_successful_operation,
            "time_since_last_success": (
                time.time() - self._last_successful_operation
                if self._last_successful_operation > 0
                else -1
            ),
        }

    async def __aenter__(self) -> SeymourClient:
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object | None,
    ) -> None:
        """Async context manager exit."""
        await self.close()

    async def _execute_operation(self, frame: bytes, receive: bool = True) -> bytes | None:
        """Execute a single transport operation with retry logic."""
        retryer = AsyncRetrying(
            stop=stop_after_attempt(self.max_retries),
            wait=wait_exponential(multiplier=0.5, min=0.5, max=10),
            retry=retry_if_exception_type(
                (SeymourTransportError, TimeoutError, ConnectionError, OSError)
            ),
            before_sleep=before_sleep_log(_LOGGER, logging.WARNING),
        )

        async for attempt in retryer:
            with attempt:
                # Ensure connection before operation
                if not self._connected:
                    await self.connect()

                # Execute with timeout
                async with asyncio.timeout(self.request_timeout):
                    await self.transport.send(frame)
                    if receive:
                        response = await self.transport.receive()
                        self._last_successful_operation = time.time()
                        return response
                    else:
                        self._last_successful_operation = time.time()
                        return None

        # This should never be reached due to retry logic, but satisfy type checker
        return None

    async def _send_and_maybe_receive(self, frame: bytes, receive: bool = True) -> bytes | None:
        """Send frame and optionally receive response with automatic reconnection."""
        try:
            logging.debug("Sending frame: %s", frame)
            return await self._execute_operation(frame, receive)
        except Exception as exc:
            # Mark disconnected for any transport error
            self._connected = False
            if isinstance(exc, SeymourTransportError | TimeoutError | ConnectionError | OSError):
                raise SeymourConnectionError("Transport operation failed after retries") from exc
            else:
                raise SeymourProtocolError("Protocol or unexpected error") from exc

    async def _send(self, frame: bytes) -> None:
        """Low-level: send frame without awaiting reply."""
        await self._send_and_maybe_receive(frame, receive=False)

    async def _send_and_receive(self, frame: bytes) -> bytes:
        """Low-level: send frame and await reply frame."""
        raw = await self._send_and_maybe_receive(frame, receive=True)
        logging.debug("Received frame: %s", raw)
        assert raw is not None
        return raw

    # Public API
    async def get_status(self) -> protocol.MaskStatus:
        raw = await self._send_and_receive(protocol.encode_status())
        return protocol.decode_status(raw)

    async def get_positions(self) -> list[protocol.MaskPosition]:
        raw = await self._send_and_receive(protocol.encode_positions())
        return protocol.decode_positions(raw)

    async def get_system_info(self) -> protocol.SystemInfo:
        raw = await self._send_and_receive(protocol.encode_read_sysinfo())
        return protocol.decode_system_info(raw)

    async def get_ratio_settings(self) -> list[protocol.RatioSetting]:
        raw = await self._send_and_receive(protocol.encode_read_settings())
        return protocol.decode_settings(raw)

    async def get_diagnostics(self, option: protocol.DiagnosticOption) -> str:
        raw = await self._send_and_receive(protocol.encode_diagnostics(option))
        return raw.decode("ascii")

    async def move_out(self, motor: protocol.MotorID, movement: protocol.MovementCode) -> None:
        frame = protocol.encode_move_out(motor, movement)
        await self._send(frame)

    async def move_in(self, motor: protocol.MotorID, movement: protocol.MovementCode) -> None:
        frame = protocol.encode_move_in(motor, movement)
        await self._send(frame)

    async def move_to_ratio(self, ratio: protocol.Ratio) -> None:
        frame = protocol.encode_move_ratio(ratio)
        await self._send(frame)

    async def home(self, motor: protocol.MotorID) -> None:
        frame = protocol.encode_home(motor)
        await self._send(frame)

    async def halt(self, motor: protocol.MotorID) -> None:
        frame = protocol.encode_halt(motor)
        await self._send(frame)

    async def calibrate(self, motor: protocol.MotorID) -> None:
        frame = protocol.encode_calibrate(motor)
        await self._send(frame)

    async def update_ratio(self, ratio: protocol.Ratio) -> None:
        frame = protocol.encode_update_ratio(ratio)
        await self._send(frame)

    async def reset_factory_default(self, ratio: protocol.Ratio | None) -> None:
        frame = protocol.encode_clear_settings(ratio)
        await self._send(frame)



"""Simple CLI utility to exercise the client."""

import asyncio
import logging
import time
from typing import Annotated

import attrs
import typer
from async_typer import AsyncTyper
from rich import print as rprint

from seymourlib.protocol import DiagnosticOption, MotorID, MovementCode, Ratio, StatusCode

from .client import SeymourClient
from .transport import SeymourTransport

app = AsyncTyper(help="Seymour controller CLI")

STATUS_POLL_INTERVAL = 0.5
STATUS_WAIT_TIMEOUT = 60.0
_MOVEMENT_COMPLETE_CODES: tuple[StatusCode, ...] = (
    StatusCode.HALTED,
    StatusCode.STOPPED_AT_RATIO,
)


@attrs.frozen
class GlobalOptions:
    verbose: bool
    host: str
    port: int
    serial_port: str | None


_GLOBAL_OPTIONS: GlobalOptions | None = None


@app.callback()  # type: ignore
def handle_global_options(
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose",
            "-v",
            help="Enable verbose output.",
        ),
    ] = False,
    host: Annotated[
        str,
        typer.Option(
            "--host",
            "-h",
            help="Hostname or IP address for TCP transport.",
            envvar="SEYMOUR_HOST",
        ),
    ] = "localhost",
    port: Annotated[
        int,
        typer.Option(
            "--port",
            "-p",
            help="Port number for TCP transport.",
            envvar="SEYMOUR_PORT",
        ),
    ] = 4999,
    serial_port: Annotated[
        str | None,
        typer.Option(
            "--serial-port",
            "-s",
            help="Serial port device for Serial transport (overrides host/port).",
            envvar="SEYMOUR_SERIAL_PORT",
        ),
    ] = None,
) -> None:
    global _GLOBAL_OPTIONS
    _GLOBAL_OPTIONS = GlobalOptions(verbose=verbose, host=host, port=port, serial_port=serial_port)
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARN)


def _get_client() -> SeymourClient:
    assert _GLOBAL_OPTIONS is not None
    if _GLOBAL_OPTIONS.serial_port:
        from .transport import SerialTransport

        transport: SeymourTransport = SerialTransport(_GLOBAL_OPTIONS.serial_port)
    else:
        from .transport import TCPTransport

        transport = TCPTransport(_GLOBAL_OPTIONS.host, _GLOBAL_OPTIONS.port)

    return SeymourClient(transport)


async def _wait_for_completion(
    client: SeymourClient,
    desired_codes: tuple[StatusCode, ...] = _MOVEMENT_COMPLETE_CODES,
    poll_interval: float = STATUS_POLL_INTERVAL,
    status_timeout: float = STATUS_WAIT_TIMEOUT,
) -> None:
    """Poll controller status until it reports one of the desired codes."""
    desired_labels = ", ".join(code.name for code in desired_codes)
    start = time.monotonic()

    # Initial delay before first poll - can take awhile for controller to report that motors are moving
    await asyncio.sleep(poll_interval * 4)

    while True:
        status = await client.get_status()
        last_status = status.code
        if last_status in desired_codes:
            assert _GLOBAL_OPTIONS is not None
            if _GLOBAL_OPTIONS.verbose:
                rprint(f"[green]Controller status:[/green] {last_status.name}")
            return

        if time.monotonic() - start >= status_timeout:
            typer.secho(
                f"Timed out waiting for status {desired_labels}; last status was {last_status.name if last_status else 'unknown'}",
                fg="red",
            )
            raise typer.Exit(code=1)

        await asyncio.sleep(poll_interval)


@app.async_command()  # type: ignore
async def status() -> None:
    """Show the current status code."""
    async with _get_client() as client:
        status = await client.get_status()
        rprint(status)


@app.async_command()  # type: ignore
async def calibrate(motor: Annotated[MotorID, typer.Argument()] = MotorID.ALL) -> None:
    """Calibrate the given motor(s) by moving all the way in & out."""
    async with _get_client() as client:
        await client.calibrate(motor)
        await _wait_for_completion(client)


positions_app = AsyncTyper(help="Commands for the motors' absolute positions.")
app.add_typer(positions_app, name="positions")


@positions_app.async_command("get")  # type: ignore
async def positions_get() -> None:
    """Show the current motor positions."""
    async with _get_client() as client:
        positions = await client.get_positions()
        for position in positions:
            rprint(position)


@positions_app.async_command("halt")  # type: ignore
async def positions_halt(motor: Annotated[MotorID, typer.Argument()] = MotorID.ALL) -> None:
    """Stop the specified motor(s) at their current position."""
    async with _get_client() as client:
        await client.halt(motor)
        await _wait_for_completion(client)


@positions_app.async_command("home")  # type: ignore
async def positions_home(motor: Annotated[MotorID, typer.Argument()] = MotorID.ALL) -> None:
    """Move the specified motor(s) to their home position."""
    async with _get_client() as client:
        await client.home(motor)
        await _wait_for_completion(client)


def _parse_increment(move: bool, jog: bool, until_limit: bool) -> MovementCode:
    match (move, jog, until_limit):
        case (True, False, False):
            return MovementCode.MOVE
        case (False, True, False):
            return MovementCode.JOG
        case (False, False, True):
            return MovementCode.UNTIL_LIMIT
        case _:
            raise typer.BadParameter(
                "Exactly one of --move, --jog, or --until-limit must be specified."
            )


@positions_app.async_command("in")  # type: ignore
async def positions_in(
    motor: Annotated[MotorID, typer.Argument()] = MotorID.ALL,
    move: bool = True,
    jog: bool = False,
    until_limit: bool = False,
) -> None:
    """Move the specified motor(s) inward by the given increment."""
    increment = _parse_increment(move, jog, until_limit)
    async with _get_client() as client:
        await client.move_in(motor, increment)
        await _wait_for_completion(client)


@positions_app.async_command("out")  # type: ignore
async def positions_out(
    motor: Annotated[MotorID, typer.Argument()] = MotorID.ALL,
    move: bool = True,
    jog: bool = False,
    until_limit: bool = False,
) -> None:
    """Move the specified motor(s) outward by the given increment."""
    increment = _parse_increment(move, jog, until_limit)
    async with _get_client() as client:
        await client.move_out(motor, increment)
        await _wait_for_completion(client)


preset_app = AsyncTyper(help="Commands for motor presets.")
app.add_typer(preset_app, name="preset")


def _parse_ratio_id(ratio_id: str) -> Ratio:
    try:
        return Ratio(ratio_id)
    except ValueError as exc:
        raise typer.BadParameter(f"Invalid ratio ID: {ratio_id}") from exc


@preset_app.async_command("apply")  # type: ignore
async def preset_apply(
    ratio_id: Annotated[str, typer.Argument(help="3-digit ratio ID, e.g. '235' for 2.35:1")],
) -> None:
    """Move motors to the designated ratio preset."""
    ratio = _parse_ratio_id(ratio_id)
    async with _get_client() as client:
        await client.move_to_ratio(ratio)
        await _wait_for_completion(client)


@preset_app.async_command("list")  # type: ignore
async def preset_list() -> None:
    """List all stored ratio presets."""
    async with _get_client() as client:
        settings = await client.get_ratio_settings()
        for preset in settings:
            rprint(preset)


@preset_app.async_command("reset")  # type: ignore
async def preset_reset(
    ratio_id: Annotated[
        str | None,
        typer.Argument(help="3-digit ratio ID, e.g. '235' for 2.35:1.  Default: ALL PRESETS"),
    ] = None,
) -> None:
    """Restore the given ratio preset(s) to their factory default."""
    ratio = _parse_ratio_id(ratio_id) if ratio_id else None
    ratio_text = ratio if ratio else "ALL PRESETS"
    typer.confirm(f"Are you sure you want to reset {ratio_text} to factory default?", abort=True)
    async with _get_client() as client:
        await client.reset_factory_default(ratio)


@preset_app.async_command("store")  # type: ignore
async def preset_store(
    ratio_id: Annotated[str, typer.Argument(help="3-digit ratio ID, e.g. '235' for 2.35:1")],
) -> None:
    """Store the current motor positions as the designated ratio preset."""
    ratio = _parse_ratio_id(ratio_id)
    async with _get_client() as client:
        await client.update_ratio(ratio)


system_app = AsyncTyper(help="Commands for reading system internals.")
app.add_typer(system_app, name="system")


@system_app.async_command("info")  # type: ignore
async def system_info() -> None:
    """Show static information about the Seymour screen system."""
    async with _get_client() as client:
        info = await client.get_system_info()
        for key, value in attrs.asdict(info).items():
            rprint(f"[bold yellow]{key}[/bold yellow]: {value}")


@system_app.async_command()  # type: ignore
async def diagnostics(option: DiagnosticOption) -> None:
    """Show diagnostic information from the Seymour screen system."""
    async with _get_client() as client:
        diag = await client.get_diagnostics(option)
        rprint(diag)


def main() -> None:
    app()
